@@ -49,6 +49,8 @@
 #define VBE_DISPI_ID0                   0xB0C0
 #define VBE_DISPI_ID1                   0xB0C1
 #define VBE_DISPI_ID2                   0xB0C2
+#define VBE_DISPI_ID3                   0xB0C3
+#define VBE_DISPI_ID4                   0xB0C4
   
 #define VBE_DISPI_DISABLED              0x00
 #define VBE_DISPI_ENABLED               0x01
@@ -83,6 +85,7 @@
     unsigned int vram_size;                                             \
     unsigned long bios_offset;                                          \
     unsigned int bios_size;                                             \
+    PCIDevice *pci_dev;                                                 \
     uint32_t latch;                                                     \
     uint8_t sr_index;                                                   \
     uint8_t sr[256];                                                    \
@@ -102,12 +105,14 @@
     uint8_t dac_read_index;                                             \
     uint8_t dac_write_index;                                            \
     uint8_t dac_cache[3]; /* used when writing */                       \
+    int dac_8bit;                                                       \
     uint8_t palette[768];                                               \
     int32_t bank_offset;                                                \
     int (*get_bpp)(struct VGAState *s);                                 \
     void (*get_offsets)(struct VGAState *s,                             \
                         uint32_t *pline_offset,                         \
-                        uint32_t *pstart_addr);                         \
+                        uint32_t *pstart_addr,                          \
+                        uint32_t *pline_compare);                       \
     void (*get_resolution)(struct VGAState *s,                          \
                         int *pwidth,                                    \
                         int *pheight);                                  \
